@@ -7,38 +7,25 @@
 #'
 #' @param resource The name of the netCDF resource to open, either a local file
 #'   name or a remote URI.
-#' @param keep_open Logical flag to indicate if the netCDF resource has to
-#'   remain open after reading the metadata. This should be enabled typically
-#'   only for programmatic access or when a remote resource has an expensive
-#'   access protocol (i.e. 2FA). The resource has to be explicitly closed with
-#'   `close()` after use. Note that when a data set is opened with
-#'   `keep_open = TRUE` the resource may still be closed by the operating system
-#'   or the remote server.
-#'
 #' @return An `CFDataset` instance, or an error if the resource was not found
 #'   or errored upon reading.
 #' @export
+#' @importFrom stats setNames
 #' @examples
 #' fn <- system.file("extdata",
 #'   "pr_day_EC-Earth3-CC_ssp245_r1i1p1f1_gr_20230101-20231231_vncdfCF.nc",
 #'   package = "ncdfCF")
 #' (ds <- open_ncdf(fn))
-open_ncdf <- function(resource, keep_open = FALSE) {
+open_ncdf <- function(resource) {
   # Parameter check
   if (length(resource) != 1L && !is.character(resource))
-    stop("Argument `resource` must be a single character string pointing to a netCDF resource.")
+    stop("Argument `resource` must be a single character string pointing to a netCDF resource.", call. = FALSE) # nocov
 
   res <- CFResource$new(resource)
-  if (is.character(res))
-    stop(res)
-
   h <- res$handle
   g <- RNetCDF::file.inq.nc(h)
 
-  nm <- regmatches(resource, regexec("([^/]*)\\.nc$", resource))[[1L]][2L]
-  if (is.na(nm))
-    nm <- regmatches(resource, regexec("([^/]*)$", resource))[[1L]][2L]
-  ds <- CFDataset$new(name = nm, resource = res, keep_open = keep_open, format = g$format)
+  ds <- CFDataset$new(resource = res, format = g$format)
 
   # Read all netCDF groups recursively
   root <- .readGroup(ds, h, vector("integer"))
@@ -49,12 +36,13 @@ open_ncdf <- function(resource, keep_open = FALSE) {
   # Find the id's of any "bounds" variables
   bnds <- sapply(root$NCvars, function(v) { # FIXME: what about dims in subgroups?
     nm <- v$attribute("bounds")
+    # FIXME: climatology
     if (!is.na(nm)) {
       obj <- v$group$find_by_name(nm, "NC")
       if (is.null(obj)) {
-        # FIXME: warning
+        warning("Unmatched `bounds` value '", nm, "' found in variable '", v$name, "'.", call. = FALSE)
         -1L
-      } else obj$dimids[1L] # By definition, bounds dimid comes first
+      } else obj$dimids[1L]    # By definition, bounds dimid comes first
     } else -1L # Flag no bounds
   })
   if (length(bnds))
@@ -68,8 +56,9 @@ open_ncdf <- function(resource, keep_open = FALSE) {
   add_dims <- all_var_dims[!(all_var_dims %in% c(all_axis_dims, bnds))]
   if (length(add_dims)) {
     axes <- append(axes, .addBareDimensions(root, add_dims))
-    names(root$CFaxes) <- sapply(root$CFaxes, function(x) x$name)
-    names(root$NCvars) <- sapply(root$NCvars, function(v) v$name)
+    axes <- axes[lengths(axes) > 0L]
+    #names(root$CFaxes) <- sapply(root$CFaxes, function(x) x$name)
+    #names(root$NCvars) <- sapply(root$NCvars, function(v) v$name)
   }
 
   # Auxiliary CVs and scalar CVs
@@ -92,7 +81,10 @@ open_ncdf <- function(resource, keep_open = FALSE) {
       "CORDEX"
     else "Generic netCDF data"
 
-    vars <- .buildVariables(root, axes)
+    # Configure any parametric vertical axes
+    .configureParametricTerms(root$CFaxes)
+
+    vars <- .buildVariables(root, root$CFaxes)
   } else {
     # Try L3b
     units <- root$attribute("units")
@@ -104,8 +96,7 @@ open_ncdf <- function(resource, keep_open = FALSE) {
         nm <- names(l3bgrp$NCvars)
         if (all(c("BinList", "BinIndex", units[1L]) %in% nm)) {
           ds$file_type <- "NASA level-3 binned data"
-          l3bgrp$CFvars <- list(CFVariableL3b$new(l3bgrp, units))
-          names(l3bgrp$CFvars) <- units[1L]
+          l3bgrp$CFvars <- setNames(list(CFVariableL3b$new(l3bgrp, units)), units[1L])
         }
       }
     }
@@ -141,10 +132,10 @@ peek_ncdf <- function(resource) {
   if (inherits(ds, "CFDataset")) {
     list(uri = ds$uri,
          type = ds$file_type,
-         variables  = do.call(rbind, lapply(ds$variables(), function(v) v$peek(grps))),
-         axes       = do.call(rbind, lapply(ds$axes(), function(a) a$peek(grps))),
+         variables  = do.call(rbind, lapply(ds$variables(), function(v) v$peek())),
+         axes       = do.call(rbind, lapply(ds$axes(), function(a) a$peek())),
          attributes = ds$attributes())
-  } else list()
+  } else list() # nocov
 }
 
 #' Read a group from a netCDF dataset
@@ -160,34 +151,29 @@ peek_ncdf <- function(resource) {
 #' @noRd
 .readGroup <- function(parent, h, parent_dims) {
   g <- RNetCDF::grp.inq.nc(h)
-  grp <- NCGroup$new(id = as.integer(g$self), name = g$name, fullname = g$fullname,
+
+  # Global attributes
+  atts <- if (g$ngatts) .readAttributes(h, "NC_GLOBAL", g$ngatts)
+          else data.frame()
+
+  grp <- NCGroup$new(id = as.integer(g$self), name = g$name, atts,
                      parent = parent, resource = parent$resource)
 
   # Read all the raw NC variables in the group
-  if (length(g$varids)) {
-    NCvars <- lapply(g$varids, function (v) .readNCVariable(grp, g$self, v))
-    names(NCvars) <- sapply(NCvars, function(v) v$name)
-    grp$NCvars <- NCvars
-  }
+  if (length(g$varids))
+    lapply(g$varids, function (v) .readNCVariable(grp, g$self, v))
 
   # Dimensions by dimid
-  if (length(g$dimids) && length(new_dims <- g$dimids[!(g$dimids %in% parent_dims)])) {
+  if (length(g$dimids) && length(new_dims <- g$dimids[!(g$dimids %in% parent_dims)]))
     dims <- lapply(new_dims, function (d) {
       dmeta <- RNetCDF::dim.inq.nc(h, d)
-      NCDimension$new(dmeta$id, dmeta$name, dmeta$length, dmeta$unlim)
+      NCDimension$new(dmeta$id, dmeta$name, dmeta$length, dmeta$unlim, grp)
     })
-    names(dims) <- sapply(dims, function(d) d$name)
-    grp$NCdims <- dims
-  }
 
   # UDTs
   if (length(g$typeids))
     grp$NCudts <- lapply(g$typeids, function(t) RNetCDF::type.inq.nc(h, t, fields = TRUE))
 
-  # Global attributes
-  if (g$ngatts)
-    grp$attributes <- .readAttributes(h, "NC_GLOBAL", g$ngatts)
-
   # Subgroups
   if (length(g$grps)) {
     grp$subgroups <- lapply(g$grps, function(z) .readGroup(grp, z, g$dimids))
@@ -201,25 +187,16 @@ peek_ncdf <- function(resource) {
 #' @noRd
 .readNCVariable <- function(grp, h, vid) {
   vmeta <- RNetCDF::var.inq.nc(h, vid)
-  var <- NCVariable$new(id = as.integer(vmeta$id), name = vmeta$name, group = grp,
-                        vtype = vmeta$type, ndims = vmeta$ndims, dimids = vmeta$dimids)
-
-  if (length(vmeta) > 6L)
-    var$netcdf4 <- vmeta[-(1L:6L)]
-
-  # Get the attributes
-  if (vmeta$natts > 0L)
-    var$attributes <- .readAttributes(h, vmeta$name, vmeta$natts)
-
-  var
+  NCVariable$new(id = as.integer(vmeta$id), name = vmeta$name, group = grp,
+                 vtype = vmeta$type, dimids = vmeta$dimids,
+                 attributes = .readAttributes(h, vmeta$name, vmeta$natts),
+                 netcdf4 = vmeta[-(1L:6L)])
 }
 
 .buildAxes <- function(grp) {
-  visible_dims <- grp$dimensions()
-
   if (length(grp$NCvars) > 0L) {
-    # Create axis for variables with name equal to dimension names
-    dim_names <- sapply(visible_dims, function(d) d$name)
+    # Create axis for local variables with name equal to visible dimensions
+    dim_names <- sapply(grp$dimensions("all"), function(d) d$name)
     if (length(dim_names)) {
       local_vars <- grp$NCvars[dim_names]
       local_CVs <- local_vars[lengths(local_vars) > 0L]
@@ -239,76 +216,71 @@ peek_ncdf <- function(resource) {
 
 # Create an `CFAxis` from an NC variable and dimension
 #
-# This method creates the various kinds of axes, with the exception of
-# [CFAxisVertical] and [CFAxisTime], which are passed off to `.makeAxisParametric`
-# once a parametric Z-axis is detected, or `.makeTimeAxis()` for a possible
-# time axis.
+# This method creates the various kinds of axes.
 #
 # @param grp Group in which the NC variable is defined.
 # @param var `NCVariable` instance to create the axis from.
 #
 # @return An instance of `CFAxis`.
+# @noRd
 .makeAxis <- function(grp, var) {
   h <- grp$handle
-  if (var$ndims == 0L)
-    dim <- NCDimension$new(-1L, var$name, 1L, FALSE) # Scalar variable
-  else
-    dim <- grp$find_dim_by_id(var$dimids[1L]) # FIXME: What about NC_CHAR axis?
 
   # Dimension values
   vals <- try(as.vector(RNetCDF::var.get.nc(h, var$name)), silent = TRUE)
   if (inherits(vals, "try-error"))
     # No dimension values so it's an identity axis
-    return(CFAxisDiscrete$new(var, dim, ""))
+    return(CFAxisDiscrete$new(var))
 
   # Does `var` have attributes?
   if (!nrow(var$attributes)) {
     # No attributes so nothing left to do
     if (var$vtype %in% c("NC_CHAR", "NC_STRING"))
-      return(CFAxisCharacter$new(var, dim, "", vals))
-    else return(CFAxisNumeric$new(var, dim, "", vals))
+      return(CFAxisCharacter$new(var, values = vals))
+    else return(CFAxisNumeric$new(var, values = vals))
   }
 
-  # Z: standard_names and formula_terms for parametric vertical axis
+  # Read some useful attributes
+  orient   <- var$attribute("axis")
   standard <- var$attribute("standard_name")
-  if (!is.na(standard) && standard %in% Z_parametric_standard_names)
-    return(.makeParametricAxis(var, dim, vals, standard))
+  units    <- var$attribute("units")
 
-  # Does the axis have bounds?
-  CFbounds <- .readBounds(grp, var$attribute("bounds"))
+  # Read the boundary values
+  bnds <- .readBounds(grp, var$attribute("bounds"))
 
-  # See if we have a "units" attribute that makes time
-  units <- var$attribute("units")
-  if (!is.na(units)) {
-    t <- .makeTimeAxis(var, units, vals)
-    if (!inherits(t, "try-error")) {
-      if (!(is.null(CFbounds)))
-        t$bounds <- CFbounds$coordinates
-      return(CFAxisTime$new(var, dim, t))
-    }
+  # See if we can make time
+  t <- .makeTimeObject(var, units, vals, bnds)
+  if (!is.null(t)) {
+    tax <- CFAxisTime$new(var, t)
+    tax$bounds <- bnds
+    return(tax)
   }
 
-  # Orientation of the axis
-  orient <- var$attribute("axis")
-  if (is.na(orient) && !is.na(units))
-    if (grepl("^degree(s?)(_?)(east|E)$", units)) orient <- "X"
-    else if (grepl("^degree(s?)(_?)(north|N)$", units)) orient <- "Y"
-  if (is.na(orient) && !is.na(standard))
-    if (standard == "longitude") orient <- "X"
-    else if (standard == "latitude") orient <- "Y"
-  if (is.na(orient)) {
-    # Desperate option: try name of NC variable - non-standard
-    nm <- toupper(var$name)
-    orient <- if (match(nm, c("X", "Y", "Z", "T"), nomatch = 0L)) nm else ""
+  # Create the axis based on units
+  axis <- if (!is.na(units)) {
+    if (grepl("^degree(s?)(_?)(east|E)$", units))
+      CFAxisLongitude$new(var, values = vals)
+    else if (grepl("^degree(s?)(_?)(north|N)$", units))
+      CFAxisLatitude$new(var, values = vals)
+    else if (requireNamespace("units", quietly = TRUE) && units::ud_are_convertible(units, "bar"))
+      CFAxisVertical$new(var, values = vals)
+    else
+      NULL
   }
+  if (is.null(axis) && !is.na(standard)) {
+    axis <- switch(standard,
+                   "longitude" = CFAxisLongitude$new(var, values = vals),
+                   "latitude"  = CFAxisLatitude$new(var, values = vals),
+                   "projection_x_coordinate" = CFAxisNumeric$new(var, values = vals, orientation = "X"),
+                   "projection_y_coordinate" = CFAxisNumeric$new(var, values = vals, orientation = "Y"),
+                   NULL)
+    if (is.null(axis) && (standard %in% Z_parametric_standard_names || !is.na(var$attribute("positive"))))
+      axis <- CFAxisVertical$new(var, values = vals)
+  }
+  if (is.na(orient)) orient <- ""
+  if (is.null(axis)) axis <- CFAxisNumeric$new(var, values = vals, orientation = orient)
 
-  axis <- if (orient == "X")
-    CFAxisLongitude$new(var, dim, vals)
-  else if (orient == "Y")
-    CFAxisLatitude$new(var, dim, vals)
-  else CFAxisNumeric$new(var, dim, orient, vals)
-  axis$bounds <- CFbounds
-
+  axis$bounds <- bnds
   axis
 }
 
@@ -326,11 +298,11 @@ peek_ncdf <- function(resource) {
   if (length(grp$NCdims) > 0L) {
     axes <- lapply(grp$NCdims, function(d) {
       if (d$id %in% add_dims) {
-        v <- NCVariable$new(-2L, d$name, grp, "NC_INT", 1L, d$id)
-        axis <- CFAxisDiscrete$new(v, d, "", dim_only = TRUE)
-        v$CF <- axis
-        grp$NCvars <- append(grp$NCvars, v)
-        grp$CFaxes <- append(grp$CFaxes, axis)
+        nm <- d$name
+        axis <- CFAxisDiscrete$new(nm, count = d$length)
+        axis$dimid <- d$id
+        lx <- list(axis); names(lx) <- nm
+        grp$CFaxes <- append(grp$CFaxes, lx)
         add_dims <- add_dims[-which(add_dims == d$id)]
         axis
       }
@@ -346,71 +318,41 @@ peek_ncdf <- function(resource) {
   axes
 }
 
-#' Make a parametric vertical axis
-#'
-#' This function is called when a parametric vertical axis is found during in
-#' function .makeAxis().
-#'
-#' @param var The NC variable that defines the axis.
-#' @param dim The NC dimension associated with the axis.
-#' @param vals The parameter values of the axis.
-#' @param param_name The "standard_name" attribute that names the specific
-#' parametric form of the axis.
-#'
-#' @return An instance of CFAxisVertical.
+#' Make a CFtime object. This will try to create a CFTime or CFClimatology object,
+#' including its bounds if set. If it fails it will return NULL, otherwise the
+#' object.
 #' @noRd
-.makeParametricAxis <- function(var, dim, vals, param_name) {
-  Z <- CFAxisVertical$new(var, dim, vals, param_name)
-
-  # Get the formula terms
-  ft <- var$attribute("formula_terms")
-  if (!is.na(ft)) {
-    ft <- trimws(strsplit(ft, " ")[[1L]], whitespace = ":")
-    dim(ft) <- c(2, length(ft) * 0.5)
-    rownames(ft) <- c("term", "variable")
-    ft <- as.data.frame(t(ft))
-    ft$NCvar <- lapply(ft$variable, function(v) {
-      ncvar <- var$group$find_by_name(v, "NC")
-      if (!is.null(ncvar)) {
-        ncvar$CF <- Z
-        ncvar
-      }
-    })
-    Z$formula_terms <- ft
-    # FIXME: Z axis may have bounds
-  }
-  Z
-}
-
-.makeTimeAxis <- function(var, units, vals) {
-  cal <- var$attribute("calendar")
-  if (is.na(cal)) cal <- "standard"
-  clim <- .readBounds(var$group, var$attribute("climatology")) # Climatology must have bounds
-  try(if (is.null(clim))
-        CFtime::CFTime$new(units, cal, vals)
-      else
-        CFtime::CFClimatology$new(units, cal, vals, clim$coordinates),
-      silent = TRUE)
+.makeTimeObject <- function(var, units, vals, bnds) {
+  if (is.na(units)) return(NULL)
+  cal <- if (is.na(cal <- var$attribute("calendar"))) "standard" else cal
+  if (!inherits(bnds, "CFBounds"))
+    bnds <- .readBounds(var$group, var$attribute("bounds"))
+  clim <- if (is.null(bnds))
+            .readBounds(var$group, var$attribute("climatology")) # Climatology must have bounds
+          else NULL
+  t <- if (is.null(clim)) try(CFtime::CFTime$new(units, cal, vals), silent = TRUE)
+  else try(CFtime::CFClimatology$new(units, cal, vals, clim$values), silent = TRUE)
+  if (inherits(t, "try-error")) return(NULL)
+  if (is.null(clim) && inherits(bnds, "CFBounds"))
+    t$bounds <- bnds$values
+  t
 }
 
 #' Make CF constructs for "coordinates" references
 #'
 #' NC variables are scanned for a "coordinates" attribute (which must be a data
-#' variable, domain variable or geometry container variable). The NC variable
-#' referenced is converted into one of 3 objects, depending on context:
-#' 1. A scalar coordinate variable in the group where its NC variable is
-#' located;
-#' 2. A label variable in the group where its NC variable is located; multiple
-#' label coordinates (such as in the case of taxon name and identifier) are
-#' stored in a single label variable;
-#' 3. A long-lat auxiliary coordinate variable when both a longitude and
-#' latitude NC variable are found, in the group of the longitude NC variable.
-#'
+#' variable, domain variable or geometry container variable). If not already
+#' present, the NC variable referenced is converted into one of 3 objects,
+#' depending on context: 1. A scalar coordinate variable in the group where its
+#' NC variable is located; 2. A label variable in the group where its NC
+#' variable is located; multiple label coordinates (such as in the case of taxon
+#' name and identifier) are stored in a single label variable; 3. A long-lat
+#' auxiliary coordinate variable when both a longitude and latitude NC variable
+#' are found, in the group of the longitude NC variable.
 #' @param grp The group to scan.
-#'
-#' @return Nothing. CFAxis and CFAuxiliaryLongLat instances are created
-#' in the groups where the NC variables are found. These will later be picked up
-#' when CFvariables are created.
+#' @return Nothing. CFAxis, CFLabel and CFAuxiliaryLongLat instances are created
+#'   in the groups where the NC variables are found. These will later be picked
+#'   up when CFVariables are created.
 #' @noRd
 .makeCoordinates <- function(grp) {
   if (length(grp$NCvars) > 0L) {
@@ -418,7 +360,7 @@ peek_ncdf <- function(resource) {
     # The NCVariable must have dimensional axes.
     for (refid in seq_along(grp$NCvars)) {
       v <- grp$NCvars[[refid]]
-      if (!length(v$CF) && length(vdimids <- v$dimids) &&
+      if (length(vdimids <- v$dimids) &&
           !is.na(coords <- v$attribute("coordinates"))) {
         coords <- strsplit(coords, " ", fixed = TRUE)[[1L]]
         varLon <- varLat <- bndsLon <- bndsLat <- NA
@@ -435,11 +377,11 @@ peek_ncdf <- function(resource) {
             if (nd == 2L && !is.na(units <- aux$attribute("units"))) {
               if (grepl("^degree(s?)(_?)(east|E)$", units)) {
                 varLon <- aux
-                bndsLon <- .readBounds(aux$group, bounds, 2L)
+                bndsLon <- .readBounds(aux$group, bounds)
                 found_one <- TRUE
               } else if (grepl("^degree(s?)(_?)(north|N)$", units)) {
                 varLat <- aux
-                bndsLat <- .readBounds(aux$group, bounds, 2L)
+                bndsLat <- .readBounds(aux$group, bounds)
                 found_one <- TRUE
               }
             }
@@ -447,24 +389,19 @@ peek_ncdf <- function(resource) {
             if (!found_one) {
               if (nd > 0L && aux$vtype %in% c("NC_CHAR", "NC_STRING")) {
                 # Label
-                dim <- grp$find_dim_by_id(aux$dimids[length(aux$dimids)]) # If there are 2 dimids, the first is a string length for a NC_CHAR type
-                val <- try(RNetCDF::var.get.nc(grp$handle, aux$name), silent = TRUE)
-                if (inherits(val, "try-error")) {
-                  warning("Could not read data for `coordinates` value '", coords[cid], "' found in variable '", v$name, "'.", call. = FALSE)
-                  next
-                }
-                aux$group$CFaux[[aux$name]] <- CFLabel$new(aux, dim, val)
+                aux$group$add_auxiliary_coordinate(CFLabel$new(aux))
                 found_one <- TRUE
               } else if (nd < 2L) {
-                # Scalar or auxiliary coordinate with a single dimension: make an axis out of it.
-                ax <- .makeAxis(grp, aux)
-                aux$group$CFaxes[[aux$name]] <- ax
-                found_one <- TRUE
+                # Scalar or auxiliary coordinate with a single dimension: make an axis out of it if it doesn't already exist.
+                if (is.null(grp$find_by_name(aux$name))) {
+                  aux$group$add_auxiliary_coordinate(.makeAxis(grp, aux))
+                  found_one <- TRUE
+                }
               }
             }
           }
 
-          if (!found_one)
+          if (!found_one && !length(aux$CF))
             warning("Unmatched `coordinates` value '", coords[cid], "' found in variable '", v$name, "'.", call. = FALSE)
         }
 
@@ -472,7 +409,13 @@ peek_ncdf <- function(resource) {
         # they have identical dimensions, in the varLon group.
         if ((inherits(varLon, "NCVariable") && inherits(varLat, "NCVariable")) &&
             identical(varLon$dimids, varLat$dimids)) {
-          varLon$group$addAuxiliaryLongLat(varLon, varLat, bndsLon, bndsLat)
+          ax <- lapply(varLon$dimids, function(did) {
+            dname <- varLon$group$find_dim_by_id(did)$name
+            varLon$group$find_by_name(dname, "NC")$CF[[1L]]
+          })
+          lonCF <- CFVariable$new(varLon, ax)
+          latCF <- CFVariable$new(varLat, ax)
+          varLon$group$add_auxiliary_longlat(lonCF, latCF, bndsLon, bndsLat)
         }
       }
     }
@@ -483,6 +426,41 @@ peek_ncdf <- function(resource) {
     lapply(grp$subgroups, function(g) .makeCoordinates(g))
 }
 
+#' @description Configure the formula terms of a parametric vertical axis. If
+#'   the vertical axis has a `formula_terms` attribute it has a parametric
+#'   coordinate space that is calculated from the formula terms. This method
+#'   sets up the axis instance to calculate the dimensional coordinate space
+#'   (but it does not do the actual calculation; access the
+#'   `parametric_coordinates` field of the vertical axis to get the dimensional
+#'   coordinates).
+#' @param axes List of `CFAxis` instances to use with the formula term objects.
+#' @return Nothing. Any parametric vertical axes in the argument `axes` will be
+#'   modified in place.
+#' @noRd
+.configureParametricTerms <- function(axes) {
+  lapply(axes, function(ax) {
+    if (!is.na(ft <- ax$attribute("formula_terms"))) {
+      ft <- trimws(strsplit(ft, " ")[[1L]], whitespace = ":")
+      dim(ft) <- c(2, length(ft) * 0.5)
+      rownames(ft) <- c("term", "variable")
+      ft <- as.data.frame(t(ft))
+      ft$param <- lapply(ft$variable, function(v) {
+        if (v == ax$name) NULL
+        else {
+          ncvar <- ax$NCvar$group$find_by_name(v, "NC")
+          if (is.null(ncvar)) {
+            CFVerticalParametricTerm$new(0, NULL)
+          } else {
+            local_axes <- .buildVariableAxisList(ncvar, axes)
+            CFVerticalParametricTerm$new(ncvar, local_axes)
+          }
+        }
+      })
+      ax$set_parametric_terms(ax$attribute("standard_name"), ft)
+    }
+  })
+}
+
 #' Make CF constructs for "cell_measures" references
 #'
 #' NC variables are scanned for a "cell_measures" attribute (which must be a
@@ -507,28 +485,32 @@ peek_ncdf <- function(resource) {
       if (!length(v$CF) && !is.na(meas <- v$attribute("cell_measures"))) {
         meas <- trimws(strsplit(meas, " ", fixed = TRUE)[[1L]], whitespace = "[ \t\r\n\\:]")
         meas <- meas[which(nzchar(meas))]
-        m <- grp$find_by_name(meas[2L], "NC")
-        if (is.null(m)) {
-          # External variable
-          root <- grp$root
-          ev <- root$attribute("external_variables")
-          if (is.na(ev) || !(meas[2L] %in% trimws(strsplit(ev, " ", fixed = TRUE)[[1L]]))) {
-            # FIXME: warning
-            warning("Unmatched `cell_measures` value '", meas[2L], "' found in variable '", v$name, "'", call. = FALSE)
-            next
+        for (m in 1:(length(meas) * 0.5)) {
+          nm <- grp$find_by_name(meas[m * 2L], "NC")
+          if (is.null(nm)) {
+            # External variable
+            root <- grp$root
+            ev <- root$attribute("external_variables")
+            if (is.na(ev) || !(meas[m * 2L] %in% trimws(strsplit(ev, " ", fixed = TRUE)[[1L]]))) {
+              # FIXME: warning
+              warning("Unmatched `cell_measures` value '", meas[m * 2L], "' found in variable '", v$name, "'", call. = FALSE)
+            } else {
+              # If it exists, move on, else create a cell measure variable
+              cmv <- grp$find_by_name(meas[m * 2L])
+              if (!inherits(cmv, "CFCellMeasure")) {
+                cm <- CFCellMeasure$new(meas[m * 2L - 1L], meas[m * 2L])
+                root$addCellMeasure(cm)
+              }
+            }
+          } else {
+            # Cell measures variable is internal. If it already exists, simply
+            # continue with the next iteration.
+            if (!length(nm$CF)) {
+              ax <- .buildVariableAxisList(nm, axes)
+              cm <- CFCellMeasure$new(meas[m * 2L - 1L], meas[m * 2L], nm, ax)
+              nm$group$addCellMeasure(cm)
+            }
           }
-          # If it exists, move on, else create a cell measure variable
-          cmv <- grp$find_by_name(meas[2L])
-          if (inherits(cmv, "CFCellMeasure")) next
-          cm <- CFCellMeasure$new(root, meas[1L], meas[2L])
-          root$addCellMeasure(cm)
-        } else {
-          # Cell measures variable is internal. If it already exists, simply
-          # continue with the next iteration.
-          if (length(m$CF)) next
-          ax <- .buildVariableAxisList(m, axes)
-          cm <- CFCellMeasure$new(m$group, meas[1L], meas[2L], m, ax)
-          m$group$addCellMeasure(cm)
         }
       }
     }
@@ -557,7 +539,7 @@ peek_ncdf <- function(resource) {
     for (refid in seq_along(grp$NCvars)) {
       v <- grp$NCvars[[refid]]
       if (!length(v$CF) && !is.na(gm <- v$attribute("grid_mapping_name")))
-        grp$CFcrs <- append(grp$CFcrs, CFGridMapping$new(v, gm))
+        grp$CFcrs <- append(grp$CFcrs, CFGridMapping$new(v))
     }
     if (length(grp$CFcrs))
       names(grp$CFcrs) <- sapply(grp$CFcrs, function(c) c$name)
@@ -570,32 +552,20 @@ peek_ncdf <- function(resource) {
 
 # Utility function to read bounds values
 # grp - the current group being processed
-# bounds - the name of the "bounds" variable, or NULL if no bounds present
-# axis_dims - number of axes on the coordinate variable; usually 1 but could be
-# 2 on an auxiliary coordinate variable
-.readBounds <- function(grp, bounds, axis_dims = 1L) {
+# bounds - the name of the boundary variable, or NA if no bounds attribute present
+.readBounds <- function(grp, bounds) {
   if (is.na(bounds)) NULL
   else {
     NCbounds <- grp$find_by_name(bounds, "NC")
     if (is.null(NCbounds)) NULL
-    else {
-      bnds <- try(RNetCDF::var.get.nc(NCbounds$group$handle, bounds, collapse = FALSE), silent = TRUE)
-      if (inherits(bnds, "try-error") || !length(bnds)) NULL
-      else {
-        if (length(dim(bnds)) == 3L && axis_dims == 1L) { # Never seen more dimensions than this
-          # FIXME: Flag non-standard item
-          bnds <- bnds[, , 1L]
-        }
-        CFBounds$new(NCbounds, grp$find_dim_by_id(NCbounds$dimids[1L]), bnds)
-      }
-    }
+    else CFBounds$new(NCbounds)
   }
 }
 
 #' Build CF variables from unused dimensional NC variables
 #'
 #' NC variables with dimensions that do not have their `CF` property set will be
-#' made into a CFVariable. This method is invoked recursively to travel through
+#' made into a `CFVariable`. This method is invoked recursively to travel through
 #' all groups of the netCDF resource.
 #'
 #' @param grp The current group to scan for unused NC variables.
@@ -606,13 +576,12 @@ peek_ncdf <- function(resource) {
   if (length(grp$NCvars) > 0L) {
     # Create variable for each unused NCVariable with dimensions
     vars <- lapply(grp$NCvars, function(v) {
+      varLon <- varLat <- ll <- NULL
       if (!length(v$CF) && v$ndims > 0L) {
-        ax <- .buildVariableAxisList(v, axes)
-        ax_names <- names(ax)
-        var <- CFVariable$new(v, ax)
+        all_ax <- dim_ax <- .buildVariableAxisList(v, axes)
+        ax_names <- names(dim_ax)
 
         # Add references to any "coordinates" of the variable
-        varLon <- varLat <- NULL
         if (!is.na(coords <- v$attribute("coordinates"))) {
           coords <- strsplit(coords, " ", fixed = TRUE)[[1L]]
           for (cid in seq_along(coords)) {
@@ -622,15 +591,16 @@ peek_ncdf <- function(resource) {
             if (!is.null(aux)) {
               clss <- class(aux)
               if (aux$length == 1L)
-                var$axes[[aux$name]] <- aux
+                all_ax[[aux$name]] <- aux
+              # FIXME: Below two branches are identical
               else if (clss[1L] == "CFLabel") {
-                ndx <- which(sapply(ax, function(x) x$dimid == aux$dimid))
-                if (length(ndx)) ax[[ndx]]$auxiliary <- aux
+                ndx <- which(sapply(dim_ax, function(x) x$dimid == aux$dimid))
+                if (length(ndx)) all_ax[[ndx]]$auxiliary <- aux
                 else {  # FIXME: record warning
                 }
               } else if ("CFAxis" %in% clss) {
-                ndx <- which(sapply(ax, function(x) x$dimid == aux$dimid))
-                if (length(ndx)) ax[[ndx]]$auxiliary <- aux
+                ndx <- which(sapply(dim_ax, function(x) x$dimid == aux$dimid))
+                if (length(ndx)) all_ax[[ndx]]$auxiliary <- aux
                 else {  # FIXME: record warning
                 }
               } else {
@@ -646,21 +616,26 @@ peek_ncdf <- function(resource) {
                 }
               }
             }
-          }
+          } # coords
 
-          if (inherits(varLon, "NCVariable") && inherits(varLat, "NCVariable")) {
+          if (inherits(varLon, "NCVariable") && inherits(varLat, "NCVariable"))
             ll <- varLon$group$find_by_name(paste(varLon$name, varLat$name, sep = "_"), "CF")
-            if (!is.null(ll)) var$gridLongLat <- ll
-          }
-        }
+        } # coordinates
+
+        # Make the CFVariable
+        var <- CFVariable$new(v, all_ax)
+        if (!is.null(ll)) var$gridLongLat <- ll
 
         # Add cell_measures
         if (!is.na(cm <- v$attribute("cell_measures"))) {
-          meas <- trimws(strsplit(cm, " ", fixed = TRUE)[[1L]][2L])
-          cmv <- v$group$find_by_name(meas, "CF")
-          if (inherits(cmv, "CFCellMeasure")) {
-            var$cell_measure <- cmv
-            cmv$register(var)
+          cms <- strsplit(cm, " ", fixed = TRUE)[[1L]]
+          len <- as.integer(length(cms) * 0.5)
+          for (i in 1L:len) {
+            cmv <- v$group$find_by_name(cms[i * 2L], "CF")
+            if (inherits(cmv, "CFCellMeasure")) {
+              var$add_cell_measure(cmv)
+              cmv$register(var)
+            }
           }
         }
 
@@ -687,7 +662,8 @@ peek_ncdf <- function(resource) {
   vars
 }
 
-#' Build a list of axes that a NC variable references.
+#' Build a list of axes that a NC variable references. These are the dimensional
+#' axes, being referenced by a dimid from the NC variable
 #'
 #' @param ncvar The NC variable to build the axis list for.
 #' @param axes List of available CF axes to use with the CF variables.
@@ -695,22 +671,25 @@ peek_ncdf <- function(resource) {
 #' @noRd
 .buildVariableAxisList <- function(ncvar, axes) {
   xids <- lapply(axes, function(x) x$dimid)
-  ax <- vector("list", ncvar$ndims)
-  for (x in 1:ncvar$ndims) {
-    ndx <- which(sapply(xids, function(e) ncvar$dimids[x] %in% e))
-    if (!length(ndx)) {
-      warning(paste0("Possible variable '", ncvar$name, "' cannot be constructed because of unknown axis identifier ", ncvar$dimids[x]))
-      # FIXME
-      return(NULL)
+  nd <- ncvar$ndims
+  if (nd > 0L) {
+    ax <- vector("list", nd)
+    for (x in 1:nd) {
+      ndx <- which(sapply(xids, function(e) ncvar$dimids[x] %in% e))
+      if (!length(ndx)) {
+        warning(paste0("Possible variable '", ncvar$name, "' cannot be constructed because of unknown axis identifier ", ncvar$dimids[x]))
+        return(NULL)
+      }
+      ax[[x]] <- axes[[ndx]]
     }
-    ax[[x]] <- axes[[ndx]]
-  }
-  names(ax) <- sapply(ax, function(x) x$name)
-  ax
+    names(ax) <- sapply(ax, function(x) x$name)
+    ax
+  } else list()
 }
 
 # Read the attributes for a group or a variable
 .readAttributes <- function(h, name, num) {
+  if (num < 1L) return(data.frame())
   atts <- do.call(rbind, lapply(0L:(num - 1L), function (a) as.data.frame(RNetCDF::att.inq.nc(h, name, a))))
   atts$value <- lapply(0L:(num - 1L), function (a) RNetCDF::att.get.nc(h, name, a, fitnum = TRUE))
   atts
